import { assertExpression, assertNotUndefined } from '../../repository/repository';
import utils from '../utils';
import { flattenMemory } from './createInstruction';
/**
 * Create assembly intructions for an assignment.
 * @returns the assembly code necessary for the assignment to happen
 */
export default function assignmentToAsm(auxVars, Left, Right, operationLine) {
    /** Main function */
    function assignmentToAsmMain() {
        switch (Left.type) {
            case 'register':
            case 'long':
            case 'fixed':
            case 'structRef':
                return leftRegularToAsm();
            case 'array':
                return leftArrayToAsm();
            default:
                throw new Error(`Internal error at line: ${operationLine}.`);
        }
    }
    /** Left side type is 'register', 'long' or 'structRef'. Create assembly instruction. */
    function leftRegularToAsm() {
        let RightMem;
        let offsetVarName;
        let assemblyCode;
        switch (Left.Offset?.type) {
            case undefined:
                switch (Right.type) {
                    case 'constant':
                        return leftRegularOffsetUndefinedAndRightConstantToAsm();
                    case 'register':
                    case 'long':
                    case 'fixed':
                    case 'structRef':
                        return leftRegularOffsetUndefinedAndRightRegularToAsm();
                    case 'array':
                        return leftRegularOffsetUndefinedAndRightArrayToAsm();
                    default:
                        throw new Error(`Internal error at line: ${operationLine}.`);
                }
            case 'constant':
                return leftRegularOffsetConstantToAsm(Left.Offset);
            case 'variable':
                RightMem = flattenMemory(auxVars, Right, operationLine);
                offsetVarName = auxVars.getMemoryObjectByLocation(Left.Offset.addr).asmName;
                assemblyCode = `SET @($${Left.asmName} + $${offsetVarName}) $${RightMem.FlatMem.asmName}\n`;
                freeIfItIsNew(RightMem);
                return RightMem.asmCode + assemblyCode;
        }
    }
    /** Left type is 'register', 'long' or ''structRef', with offset undefined. Right type is 'constant'.
     * Create assembly instruction. */
    function leftRegularOffsetUndefinedAndRightConstantToAsm() {
        let newVarName;
        switch (Right.Offset?.type) {
            case undefined:
                return leftRegularOffsetUndefinedAndRightConstantOffsetUndefinedToAsm();
            case 'constant':
                Right.hexContent = assertNotUndefined(Right.hexContent);
                newVarName = auxVars.getMemoryObjectByLocation(utils.addHexSimple(Right.Offset.value, Right.hexContent), operationLine).asmName;
                return `SET @${Left.asmName} $${newVarName}\n`;
            case 'variable':
                throw new Error('Not implemented.');
        }
    }
    function leftRegularOffsetUndefinedAndRightConstantOffsetUndefinedToAsm() {
        Right.hexContent = assertNotUndefined(Right.hexContent);
        if (Right.hexContent.length > 17) {
            throw new Error(`At line: ${operationLine}.` +
                ' Overflow on long value assignment (value bigger than 64 bits)');
        }
        if (Right.hexContent === '0000000000000000') {
            return `CLR @${Left.asmName}\n`;
        }
        let optVarName = 'n';
        if (Right.declaration === 'fixed') {
            optVarName = 'f';
        }
        optVarName += Number('0x' + Right.hexContent).toString(10);
        const findOpt = auxVars.memory.find(MEM => MEM.asmName === optVarName && MEM.hexContent === Right.hexContent);
        if (findOpt) {
            return `SET @${Left.asmName} $${findOpt.asmName}\n`;
        }
        return `SET @${Left.asmName} #${Right.hexContent}\n`;
    }
    /** Left type is 'register', 'long' or 'structRef', with offset undefined. Right type is 'register', 'long' or
     * 'structRef'. Create assembly instruction. */
    function leftRegularOffsetUndefinedAndRightRegularToAsm() {
        let offsetVarName;
        switch (Right.Offset?.type) {
            case undefined:
                return leftRegularOffsetUndefinedAndRightRegularOffsetUndefinedToAsm();
            case 'constant':
                return leftRegularOffsetUndefinedAndRightRegularOffsetConstantToAsm(Right.Offset);
            case 'variable':
                offsetVarName = auxVars.getMemoryObjectByLocation(Right.Offset.addr, operationLine).asmName;
                return `SET @${Left.asmName} $($${Right.asmName} + $${offsetVarName})\n`;
        }
    }
    /** Left type is 'register', 'long' or 'structRef', with offset undefined. Right type is 'register', 'long', or
     * 'structRef' with offset undefined. Create assembly instruction. */
    function leftRegularOffsetUndefinedAndRightRegularOffsetUndefinedToAsm() {
        if (utils.isNotValidDeclarationOp(Left.declaration, Right)) {
            throw new Error(`Internal error at line: ${operationLine}.`);
        }
        if (Left.address === Right.address) {
            return '';
        }
        return `SET @${Left.asmName} $${Right.asmName}\n`;
    }
    /** Left type is 'register', 'long' or 'structRef', with offset undefined. Right type is 'register', 'long' or
     * 'structRef' with offset constant. Create assembly instruction. */
    function leftRegularOffsetUndefinedAndRightRegularOffsetConstantToAsm(RightOffset) {
        if (RightOffset.value === 0) {
            return `SET @${Left.asmName} $($${Right.asmName})\n`;
        }
        const MemOffset = flattenMemory(auxVars, utils.createConstantMemObj(RightOffset.value), operationLine);
        const assemblyCode = `SET @${Left.asmName} $($${Right.asmName} + $${MemOffset.FlatMem.asmName})\n`;
        freeIfItIsNew(MemOffset);
        return MemOffset.asmCode + assemblyCode;
    }
    /** Left type is 'register', 'long' or 'structRef', with offset undefined. Right type is 'array'.
     * Create assembly instruction. */
    function leftRegularOffsetUndefinedAndRightArrayToAsm() {
        if (Right.Offset === undefined) {
            return `SET @${Left.asmName} $${Right.asmName}\n`;
        }
        if (Right.Offset.type === 'constant') {
            const memLoc = utils.addHexSimple(Right.hexContent, Right.Offset.value);
            const RightMem = auxVars.getMemoryObjectByLocation(memLoc, operationLine);
            return `SET @${Left.asmName} $${RightMem.asmName}\n`;
        }
        // Right.Offset.type is 'variable'
        const offsetVarName = auxVars.getMemoryObjectByLocation(Right.Offset.addr, operationLine).asmName;
        return `SET @${Left.asmName} $($${Right.asmName} + $${offsetVarName})\n`;
    }
    /** Left type is 'register', 'long' or 'structRef', with offset constant. Create assembly instruction. */
    function leftRegularOffsetConstantToAsm(LeftOffset) {
        const RightMem = flattenMemory(auxVars, Right, operationLine);
        let assemblyCode;
        if (LeftOffset.value === 0) {
            assemblyCode = `SET @($${Left.asmName}) $${RightMem.FlatMem.asmName}\n`;
            if (RightMem.isNew) {
                auxVars.freeRegister(RightMem.FlatMem.address);
            }
            return RightMem.asmCode + assemblyCode;
        }
        const MemOffset = flattenMemory(auxVars, utils.createConstantMemObj(LeftOffset.value), operationLine);
        assemblyCode = `SET @($${Left.asmName} + $${MemOffset.FlatMem.asmName}) $${RightMem.FlatMem.asmName}\n`;
        freeIfItIsNew(MemOffset);
        freeIfItIsNew(RightMem);
        return RightMem.asmCode + MemOffset.asmCode + assemblyCode;
    }
    /** Left type is 'array'. Create assembly instruction. */
    function leftArrayToAsm() {
        let RightMem;
        let assemblyCode;
        let leftOffsetVarName;
        switch (Left.Offset?.type) {
            case undefined:
                return leftArrayOffsetUndefinedToAsm();
            case 'constant':
                // Optimimization steps before lead to impossible reach code
                throw new Error(`Internal error at line: ${operationLine}.`);
            case 'variable':
                RightMem = flattenMemory(auxVars, Right, operationLine);
                leftOffsetVarName = auxVars.getMemoryObjectByLocation(Left.Offset.addr, operationLine).asmName;
                assemblyCode = `SET @($${Left.asmName} + $${leftOffsetVarName}) $${RightMem.FlatMem.asmName}\n`;
                freeIfItIsNew(RightMem);
                return RightMem.asmCode + assemblyCode;
        }
    }
    /** Left type is 'array', with offset undefined. Create assembly instruction. */
    function leftArrayOffsetUndefinedToAsm() {
        assertExpression(Right.type === 'constant', `Internal error at line: ${operationLine}.`);
        // special case for multi-long text assignment
        const arraySize = assertNotUndefined(Left.ArrayItem).totalSize - 1;
        if (Right.size > arraySize) {
            throw new Error(`At line: ${operationLine}.` +
                ' Overflow on array value assignment (value bigger than array size).');
        }
        const paddedLong = assertNotUndefined(Right.hexContent).padStart(arraySize * 16, '0');
        let assemblyCode = '';
        for (let i = 0; i < arraySize; i++) {
            const newLeft = auxVars.getMemoryObjectByLocation(utils.addHexSimple(Left.hexContent, i), operationLine);
            const newRight = utils.createConstantMemObj(paddedLong.slice(16 * (arraySize - i - 1), 16 * (arraySize - i)));
            assemblyCode += assignmentToAsm(auxVars, newLeft, newRight, operationLine);
        }
        return assemblyCode;
    }
    function freeIfItIsNew(FlatObj) {
        if (FlatObj.isNew) {
            auxVars.freeRegister(FlatObj.FlatMem.address);
        }
    }
    return assignmentToAsmMain();
}
