import { stringToHexstring, ReedSalomonAddressDecode, parseDecimalNumber } from '../repository/repository';
/** Translate an array of pre tokens to an array of tokens. First phase of parsing.
 * @param tokens Array of pre-tokens
 * @returns Array of TOKENS. Recursive on Arr, CodeCave and CodeDomain types
 * @throws {Error} at any mistakes
 */
export default function parser(preTokens) {
    // This object stores a recipe to transform one or more pre_tokens into one
    //   token. All non-recursive items are here. The order they are
    //   arrange are important to decide in cases where same element token can be
    //   more than one functional token depending the other tokens after it.
    // Recursive tokens will be treated on getNextToken() function.
    const notRecursiveTokensSpecs = [
        // single-tokens easy
        {
            sequence: ['tilde'],
            action(tokenID) {
                return { type: 'UnaryOperator', precedence: 2, value: '~', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['semi'],
            action(tokenID) {
                return { type: 'Terminator', precedence: 12, value: ';', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['comma'],
            action(tokenID) {
                return { type: 'Delimiter', precedence: 11, value: ',', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['colon'],
            action(tokenID) {
                return { type: 'Colon', precedence: 0, value: ':', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['dot'],
            action(tokenID) {
                return { type: 'Member', precedence: 0, value: '.', line: preTokens[tokenID].line };
            }
        },
        // single-tokens medium
        {
            sequence: ['macro'],
            action(tokenID) {
                return { type: 'Macro', precedence: 0, value: preTokens[tokenID].value, line: preTokens[tokenID].line };
            }
        },
        // single-tokens hard
        {
            sequence: ['keyword'],
            action(tokenID) {
                const Node = {
                    type: 'Keyword',
                    precedence: 12,
                    value: preTokens[tokenID].value,
                    line: preTokens[tokenID].line
                };
                if (preTokens[tokenID].value === 'asm' || preTokens[tokenID].value === 'struct') {
                    Node.extValue = preTokens[tokenID].extValue;
                }
                if (preTokens[tokenID].value === 'sizeof') {
                    Node.precedence = 2;
                }
                return Node;
            }
        },
        {
            sequence: ['numberDec'],
            action(tokenID) {
                const PreTkn = preTokens[tokenID];
                const Parsed = parseDecimalNumber(PreTkn.value, PreTkn.line);
                const valString = Parsed.value.toString(16);
                const paddedValString = valString.padStart((Math.floor((valString.length - 1) / 16) + 1) * 16, '0');
                return { type: 'Constant', precedence: 0, value: paddedValString, line: PreTkn.line, extValue: Parsed.declaration };
            }
        },
        {
            sequence: ['numberHex'],
            action(tokenID) {
                const PreTkn = preTokens[tokenID];
                let val = PreTkn.value.replace(/_/g, '').toLowerCase();
                val = val.padStart((Math.floor((val.length - 1) / 16) + 1) * 16, '0');
                return { type: 'Constant', precedence: 0, value: val, line: PreTkn.line, extValue: 'long' };
            }
        },
        {
            sequence: ['string'],
            action(tokenID) {
                const PreTkn = preTokens[tokenID];
                let val = stringToHexstring(PreTkn.value);
                const parts = /^(BURST-|S-|TS-)([0-9A-Z]{4}-[0-9A-Z]{4}-[0-9A-Z]{4}-[0-9A-Z]{5})/.exec(PreTkn.value);
                if (parts !== null) {
                    val = ReedSalomonAddressDecode(parts[2], PreTkn.line);
                }
                return { type: 'Constant', precedence: 0, value: val, line: PreTkn.line };
            }
        },
        // multi-tokens easy
        {
            sequence: ['equal', 'equal'],
            action(tokenID) {
                return { type: 'Comparision', precedence: 6, value: '==', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['equal'],
            action(tokenID) {
                return { type: 'Assignment', precedence: 10, value: '=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['not', 'equal'],
            action(tokenID) {
                return { type: 'Comparision', precedence: 6, value: '!=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['not'],
            action(tokenID) {
                return { type: 'UnaryOperator', precedence: 2, value: '!', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['forwardslash', 'equal'],
            action(tokenID) {
                return { type: 'SetOperator', precedence: 10, value: '/=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['forwardslash'],
            action(tokenID) {
                return { type: 'Operator', precedence: 3, value: '/', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['percent', 'equal'],
            action(tokenID) {
                return { type: 'SetOperator', precedence: 10, value: '%=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['percent'],
            action(tokenID) {
                return { type: 'Operator', precedence: 3, value: '%', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['less', 'less', 'equal'],
            action(tokenID) {
                return { type: 'SetOperator', precedence: 10, value: '<<=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['less', 'less'],
            action(tokenID) {
                return { type: 'Operator', precedence: 5, value: '<<', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['less', 'equal'],
            action(tokenID) {
                return { type: 'Comparision', precedence: 6, value: '<=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['less'],
            action(tokenID) {
                return { type: 'Comparision', precedence: 6, value: '<', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['pipe', 'pipe'],
            action(tokenID) {
                return { type: 'Comparision', precedence: 9, value: '||', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['pipe', 'equal'],
            action(tokenID) {
                return { type: 'SetOperator', precedence: 10, value: '|=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['pipe'],
            action(tokenID) {
                return { type: 'Operator', precedence: 7, value: '|', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['greater', 'equal'],
            action(tokenID) {
                return { type: 'Comparision', precedence: 6, value: '>=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['greater', 'greater', 'equal'],
            action(tokenID) {
                return { type: 'SetOperator', precedence: 10, value: '>>=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['greater', 'greater'],
            action(tokenID) {
                return { type: 'Operator', precedence: 5, value: '>>', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['greater'],
            action(tokenID) {
                return { type: 'Comparision', precedence: 6, value: '>', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['caret', 'equal'],
            action(tokenID) {
                return { type: 'SetOperator', precedence: 10, value: '^=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['caret'],
            action(tokenID) {
                return { type: 'Operator', precedence: 7, value: '^', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['variable'],
            action(tokenID) {
                return {
                    type: 'Variable',
                    precedence: 0,
                    value: preTokens[tokenID].value,
                    line: preTokens[tokenID].line
                };
            }
        },
        // multi-tokens medium
        {
            sequence: ['star', 'equal'],
            action(tokenID) {
                return { type: 'SetOperator', precedence: 10, value: '*=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['star'],
            action(tokenID) {
                if (isBinaryOperator(tokenID)) {
                    return { type: 'Operator', precedence: 3, value: '*', line: preTokens[tokenID].line };
                }
                return { type: 'UnaryOperator', precedence: 2, value: '*', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['plus', 'equal'],
            action(tokenID) {
                return { type: 'SetOperator', precedence: 10, value: '+=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['plus', 'plus'],
            action(tokenID) {
                return { type: 'SetUnaryOperator', precedence: 1, value: '++', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['plus'],
            action(tokenID) {
                if (isBinaryOperator(tokenID)) {
                    return { type: 'Operator', precedence: 4, value: '+', line: preTokens[tokenID].line };
                }
                return { type: 'UnaryOperator', precedence: 2, value: '+', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['minus', 'minus'],
            action(tokenID) {
                return { type: 'SetUnaryOperator', precedence: 1, value: '--', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['minus', 'equal'],
            action(tokenID) {
                return { type: 'SetOperator', precedence: 10, value: '-=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['minus', 'greater'],
            action(tokenID) {
                return { type: 'Member', precedence: 0, value: '->', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['minus'],
            action(tokenID) {
                if (isBinaryOperator(tokenID)) {
                    return { type: 'Operator', precedence: 4, value: '-', line: preTokens[tokenID].line };
                }
                return { type: 'UnaryOperator', precedence: 2, value: '-', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['and', 'and'],
            action(tokenID) {
                return { type: 'Comparision', precedence: 8, value: '&&', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['and', 'equal'],
            action(tokenID) {
                return { type: 'SetOperator', precedence: 10, value: '&=', line: preTokens[tokenID].line };
            }
        },
        {
            sequence: ['and'],
            action(tokenID) {
                if (isBinaryOperator(tokenID)) {
                    return { type: 'Operator', precedence: 7, value: '&', line: preTokens[tokenID].line };
                }
                return { type: 'UnaryOperator', precedence: 2, value: '&', line: preTokens[tokenID].line };
            }
        }
    ];
    const AuxVars = {
        mainLoopIndex: 0
    };
    // Will run once
    function parseMain() {
        const tokenTrain = [];
        while (AuxVars.mainLoopIndex < preTokens.length) {
            tokenTrain.push(getNextToken());
        }
        return tokenTrain;
    }
    // Process element preTokens started at position mainLoopIndex (outer scope) and returns a functional token
    function getNextToken() {
        const CurrentPreToken = preTokens[AuxVars.mainLoopIndex];
        let RetToken;
        // take care of not recursive tokens
        const FoundRule = notRecursiveTokensSpecs.find(matchRule);
        if (FoundRule !== undefined) {
            RetToken = FoundRule.action(AuxVars.mainLoopIndex);
            AuxVars.mainLoopIndex += FoundRule.sequence.length;
            return RetToken;
        }
        // take care of recursive tokens
        switch (CurrentPreToken.value) {
            case ']':
            case ')':
            case '}':
                throw new Error(`At line: ${CurrentPreToken.line}. Unexpected closing '${CurrentPreToken.value}'.`);
            case '[':
                RetToken = { type: 'Arr', value: '', precedence: 0, line: CurrentPreToken.line };
                AuxVars.mainLoopIndex++;
                RetToken.params = getTokensUntil(']', RetToken.type, RetToken.line);
                return RetToken;
            case '(':
                if (AuxVars.mainLoopIndex > 0 && preTokens[AuxVars.mainLoopIndex - 1].type === 'variable') {
                    RetToken = { type: 'Function', value: '', precedence: 0, line: CurrentPreToken.line };
                }
                else {
                    RetToken = { type: 'CodeCave', value: '', precedence: 0, line: CurrentPreToken.line };
                }
                AuxVars.mainLoopIndex++;
                RetToken.params = getTokensUntil(')', RetToken.type, RetToken.line);
                return RetToken;
            case '{':
                RetToken = { type: 'CodeDomain', value: '', precedence: 0, line: CurrentPreToken.line };
                AuxVars.mainLoopIndex++;
                RetToken.params = getTokensUntil('}', RetToken.type, RetToken.line);
                return RetToken;
            default:
                throw new Error(`At line: ${CurrentPreToken.line}.` +
                    ` Unexpected token '${CurrentPreToken.value}' - type: '${CurrentPreToken.type}'.`);
        }
    }
    // Process element preTokens started at position mainLoopIndex (outer scope) and returns a array of tokens
    // until endChar is found
    function getTokensUntil(endChar, parentType, line) {
        const returnedTokens = [];
        if (AuxVars.mainLoopIndex >= preTokens.length) {
            throw new Error('At line: end of file. ' +
                `Missing closing '${endChar}' for for '${parentType}' started at line: ${line}.`);
        }
        while (preTokens[AuxVars.mainLoopIndex].value !== endChar) {
            returnedTokens.push(getNextToken());
            // getNextToken will increase mainLoopIndex for loop
            if (AuxVars.mainLoopIndex >= preTokens.length) {
                throw new Error('At line: end of file. ' +
                    `Missing closing '${endChar}' for for '${parentType}' started at line: ${line}.`);
            }
        }
        // discard closing char
        AuxVars.mainLoopIndex++;
        return returnedTokens;
    }
    function matchRule(RuleN) {
        for (let i = 0; i < RuleN.sequence.length; i++) {
            if (preTokens[AuxVars.mainLoopIndex + i]?.type === RuleN.sequence[i])
                continue;
            return false; // proceed to next rule
        }
        return true; // all sequence matched!
    }
    // Use to detect if a token at some position is Binary or Unary
    function isBinaryOperator(position) {
        if (position >= 2) {
            if ((preTokens[position - 1].type === 'plus' && preTokens[position - 2].type === 'plus') ||
                (preTokens[position - 1].type === 'minus' && preTokens[position - 2].type === 'minus')) {
                return true;
            }
        }
        if (position >= 1) {
            if (preTokens[position - 1].type === 'variable' ||
                preTokens[position - 1].type === 'numberDec' ||
                preTokens[position - 1].type === 'numberHex' ||
                preTokens[position - 1].type === 'string' ||
                preTokens[position - 1].value === ']' ||
                preTokens[position - 1].value === ')') {
                return true;
            }
        }
        return false;
    }
    return parseMain();
}
