import { assertExpression } from '../repository/repository';
/**
 * Transforms inputSourceCode into an array of pre tokens.
 * This array is not recursive.
 * @param inputSourceCode source code text
 * @returns array of pre tokens
 */
export default function tokenizer(inputSourceCode) {
    const simpleTokensSpecs = [
        { char: '=', pretokenType: 'equal' },
        { char: '*', pretokenType: 'star' },
        { char: '!', pretokenType: 'not' },
        { char: '[', pretokenType: 'bracket' },
        { char: ']', pretokenType: 'bracket' },
        { char: '-', pretokenType: 'minus' },
        { char: '+', pretokenType: 'plus' },
        { char: '\\', pretokenType: 'backslash' },
        { char: '/', pretokenType: 'forwardslash' },
        { char: '.', pretokenType: 'dot' },
        { char: '<', pretokenType: 'less' },
        { char: '>', pretokenType: 'greater' },
        { char: '|', pretokenType: 'pipe' },
        { char: '&', pretokenType: 'and' },
        { char: '%', pretokenType: 'percent' },
        { char: '^', pretokenType: 'caret' },
        { char: ',', pretokenType: 'comma' },
        { char: ';', pretokenType: 'semi' },
        { char: '~', pretokenType: 'tilde' },
        { char: '`', pretokenType: 'grave' },
        { char: '(', pretokenType: 'paren' },
        { char: ')', pretokenType: 'paren' },
        { char: ':', pretokenType: 'colon' },
        { char: '{', pretokenType: 'curly' },
        { char: '}', pretokenType: 'curly' },
        { char: '#', pretokenType: 'SPECIAL' }
    ];
    const regexSingleTokensSpecs = [
        {
            start: /^(\/\/.*)/,
            pretokenType: 'NONE',
            addLength: 0
        },
        {
            start: /^(\s+)/,
            pretokenType: 'NONE',
            addLength: 0
        },
        {
            start: /(^\d[\d_.]*\b|^\.[\d_]+\b)/,
            pretokenType: 'numberDec',
            addLength: 0
        },
        {
            start: /^0[xX]([\da-fA-F][\da-fA-F_]*\b)/,
            pretokenType: 'numberHex',
            addLength: 2
        },
        {
            start: /^(break|case|const|continue|default|do|else|exit|fixed|for|goto|halt|if|inline|long|return|sleep|sizeof|switch|void|while)\b/,
            pretokenType: 'keyword',
            addLength: 0
        },
        {
            start: /^(asm)/,
            pretokenType: 'ASM',
            addLength: 0
        },
        {
            start: /^(struct)/,
            pretokenType: 'STRUCT',
            addLength: 0
        },
        {
            start: /^(\w+)/,
            pretokenType: 'variable',
            addLength: 0
        }
    ];
    const regexDoubleTokensSpecs = [
        {
            start: /^\/\*/,
            end: /^([\s\S]*?\*\/)/,
            pretokenType: 'NONE',
            startLength: 2,
            removeTrailing: 0,
            errorMsg: "Missing '*/' to end comment section."
        },
        {
            start: /^"/,
            end: /^([\s\S]*?")/,
            pretokenType: 'string',
            startLength: 1,
            removeTrailing: 1,
            errorMsg: "Missing '\"' to end string."
        },
        {
            start: /^'/,
            end: /^([\s\S]*?')/,
            pretokenType: 'string',
            startLength: 1,
            removeTrailing: 1,
            errorMsg: "Missing \"'\" to end string."
        }
    ];
    const AuxVars = {
        currentChar: '',
        remainingText: '',
        current: 0,
        preTokens: [],
        currentLine: 1
    };
    function tokenizeMain() {
        while (AuxVars.current < inputSourceCode.length) {
            AuxVars.currentChar = inputSourceCode.charAt(AuxVars.current);
            AuxVars.remainingText = inputSourceCode.slice(AuxVars.current);
            // Resolve double regex preTokens
            if (regexDoubleTokensSpecs.find(findAndProcessDoubleTokens)) {
                continue;
            }
            // Resolve single regex preTokens
            if (regexSingleTokensSpecs.find(findAndProcessSingleTokens)) {
                continue;
            }
            // Resolve all simple preTokens
            if (simpleTokensSpecs.find(findAndProcessSimpleTokens)) {
                continue;
            }
            throw new Error(`At line: ${AuxVars.currentLine}.` +
                ` Forbidden character found: '${AuxVars.currentChar}'.`);
        }
        return AuxVars.preTokens;
    }
    function findAndProcessDoubleTokens(RuleDouble) {
        const startParts = RuleDouble.start.exec(AuxVars.remainingText);
        if (startParts === null) {
            return false;
        }
        const endParts = RuleDouble.end.exec(AuxVars.remainingText.slice(RuleDouble.startLength));
        AuxVars.current += RuleDouble.startLength;
        if (endParts === null) {
            throw new Error(`At line: ${AuxVars.currentLine}. ${RuleDouble.errorMsg}`);
        }
        if (RuleDouble.pretokenType === 'NONE') {
            AuxVars.currentLine += (endParts[1].match(/\n/g) ?? '').length;
            AuxVars.current += endParts[1].length;
            return true; // breaks find function
        }
        AuxVars.preTokens.push({
            type: RuleDouble.pretokenType,
            value: endParts[1].slice(0, -RuleDouble.removeTrailing),
            line: AuxVars.currentLine
        });
        AuxVars.currentLine += (endParts[1].match(/\n/g) ?? '').length;
        AuxVars.current += endParts[1].length;
        return true; // breaks find function
    }
    function findAndProcessSingleTokens(RuleSingle) {
        const startParts = RuleSingle.start.exec(AuxVars.remainingText);
        if (startParts === null) {
            return false;
        }
        switch (RuleSingle.pretokenType) {
            case 'NONE':
                AuxVars.currentLine += (startParts[1].match(/\n/g) ?? '').length;
                AuxVars.current += startParts[1].length + RuleSingle.addLength;
                return true;
            case 'ASM': {
                const asmParts = /^(asm[^\w]*\{)([\s\S]*)/.exec(AuxVars.remainingText);
                if (asmParts === null) {
                    throw new Error(`At line: ${AuxVars.currentLine}. Error parsing 'asm { ... }' keyword`);
                }
                const endLocation = asmParts[2].indexOf('}');
                if (endLocation === -1) {
                    throw new Error(`At line: ${AuxVars.currentLine}.` +
                        " Ending '}' not found for 'asm { ... }' keyword.");
                }
                const asmText = asmParts[2].slice(0, endLocation);
                const asmCode = asmParts[1] + asmText + '}';
                AuxVars.preTokens.push({ type: 'keyword', value: 'asm', line: AuxVars.currentLine, extValue: asmText });
                AuxVars.currentLine += (asmCode.match(/\n/g) ?? '').length;
                AuxVars.current += asmCode.length;
                return true;
            }
            case 'STRUCT': {
                const structParts = /^(struct\s+(\w+))/.exec(AuxVars.remainingText);
                if (structParts === null) {
                    throw new Error(`At line: ${AuxVars.currentLine}.` +
                        " 'struct' keyword must be followed by a type name");
                }
                AuxVars.preTokens.push({
                    type: 'keyword',
                    value: 'struct',
                    line: AuxVars.currentLine,
                    extValue: structParts[2]
                });
                AuxVars.currentLine += (structParts[1].match(/\n/g) ?? '').length;
                AuxVars.current += structParts[1].length;
                return true;
            }
            default:
                AuxVars.preTokens.push({ type: RuleSingle.pretokenType, value: startParts[1], line: AuxVars.currentLine });
                AuxVars.currentLine += (startParts[1].match(/\n/g) ?? '').length;
                AuxVars.current += startParts[1].length + RuleSingle.addLength;
                return true;
        }
    }
    function findAndProcessSimpleTokens(SimpleItem) {
        if (SimpleItem.char !== AuxVars.currentChar) {
            return false;
        }
        if (SimpleItem.pretokenType === 'SPECIAL') {
            assertExpression(SimpleItem.char === '#', 'Internal error at tokenizer.');
            AuxVars.current++;
            const val = inputSourceCode.slice(AuxVars.current).split('\n')[0];
            AuxVars.current += val.length + 1; // newline!
            AuxVars.currentLine++;
            AuxVars.preTokens.push({ type: 'macro', value: val, line: AuxVars.currentLine - 1 });
            return true;
        }
        AuxVars.preTokens.push({ type: SimpleItem.pretokenType, value: AuxVars.currentChar, line: AuxVars.currentLine });
        AuxVars.current++;
        return true;
    }
    return tokenizeMain();
}
